//------------------------------------------------------------------------------
// Прошивка для датчика RMS
// Код семейства - 0xA1
// Адрес на шине i2c - 0x11
// OldBean, 28.11.2018
//------------------------------------------------------------------------------
#include <avr/io.h>
#include <util/twi.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <string.h>
//------------------------------------------------------------------------------
#define F_CPU 16000000UL // Тактовая частота МК
#include <util/delay.h>

#define ADDR 0x11        // Адрес устройства на шине I2C
#define FAMILY_CODE 0xA1 // Код семейства данного устройства
#define BUF_SIZE 32      // Размер буфера обмена данными
#define EE_SIZE 6        // Размер блока идентиф. и калибр. данных
//------------------------------------------------------------------------------
#define ZERO_START 106 // Определяет задержку фронта импульса Zero
#define ZERO_WIDTH 13  // Определяет ширину импульса Zero (100 мкс)
#define T_ADC 49       // Определяет интервал между запусками АЦП (200 мкс)
#define SUM_MAX 5000   // Количество суммируемых отсчетов (5000 - это за 1 сек)

static volatile uint32_t sum;      // Сумматор квадратов отсчетов АЦП
static volatile uint32_t crms;     // Сумма квадратов отсчетов АЦП для хранения
static volatile uint16_t sc = 0;   // Счетчик просуммированных квадратов
static volatile uint8_t zFlag = 0; // Разрешение формирования импульсов Zero
//------------------------------------------------------------------------------
static uint8_t buf[BUF_SIZE];        // Буфер обмена
static uint8_t scratch[BUF_SIZE];    // Рабочая область RAM
static volatile uint8_t *ptr;        // Указатель на байт буфера обмена
static volatile uint8_t runFlag = 1; // Рабочий режим, 0 - режим настройки
static volatile uint8_t nb = 0;      // Счетчик байтов
static volatile uint8_t prFlag = 0;  // Принят новый пакет байтов
static volatile uint8_t cmd = 0x00;  // Код команды
//------------------------------------------------------------------------------
ISR(TWI_vect) { // Обработчик событий модуля TWI
  switch(TW_STATUS) {     // Статус - в 5 старших битах регистра TWSR
    // Принимаем байты
    case TW_SR_SLA_ACK:   // Адрес для приема принят, ACK отправлен - 0x80
      ptr = buf;          // Настраиваемся на начало буфера обмена
      nb = 0;             // Сброс счетчика байтов
      break;
    case TW_SR_DATA_ACK:  // Очередной байт принят, ACK отправлен - 0x80
      *ptr++ = TWDR;
      nb++;
      break;
    case TW_SR_STOP:      // Состояние "стоп" на шине - 0xA0
      prFlag = 1;         // Новый пакет принят
      break;
    // Отправляем байты
    case TW_ST_SLA_ACK:   // Адрес для отправки принят, ACK отправлен - 0xA8
      ptr = buf;          // Настраиваемся на начало буфера обмена
      nb = 0;             // Сброс счетчика байтов
    case TW_ST_DATA_ACK:  // Очередной байт отправлен, ACK принят - 0xB8
      TWDR = *ptr++;
      nb++;
      break;
    case TW_ST_DATA_NACK: // Байт отправлен и получен NACK - 0xC0
    case TW_ST_LAST_DATA: // Отправлен последний байт, получен ACK - 0xC8
      break;
    // Ошибки старта или стопа - сброс TWI
    case TW_BUS_ERROR:
      TWCR = _BV(TWIE) | _BV(TWINT) | _BV(TWEA) | _BV(TWEN);
      break;
  }
  TWCR |= _BV(TWINT); // Очистим флаг TWINT (очищается при установке в 1)
}
//------------------------------------------------------------------------------
ISR(TIMER0_COMPA_vect) { // Обработчик по совпадению с регистром A
  ADCSRA |= _BV(ADSC); // Запуск преобразования
}
//------------------------------------------------------------------------------
ISR(ADC_vect) { // Преобразование закончено
  register uint32_t val = ADCL; val |= ((uint32_t)ADCH) << 8;
  sei(); // Следующие фрагменты не требуют атомарности; разрешим прерывания
  sum += val*val; // Суммирование квадратов кода АЦП - 6 мкс
  if(++sc == SUM_MAX) { // Насуммировали достаточно
    crms = sum;
    sc = 0; sum = 0; // Сбрасываем счетчик и сумматор
  }
}
//------------------------------------------------------------------------------
ISR(INT0_vect) { // Внешнее прерывание на пине PB2/INT0
  GTCCR = _BV(PSRASY); // Сброс предделителя таймера 2
  TCNT2 = 0; // Обнуляем счетчик таймера 2
  zFlag = 1; // Разрешаем формирование импульса Zero
}
//------------------------------------------------------------------------------
ISR(TIMER2_COMPA_vect) { // Обработчик по совпадению с регистром A
  if(!zFlag) return;
  PORTB |= _BV(PB0); // Формируем фронт импульса на шине Zero
}
//------------------------------------------------------------------------------
ISR(TIMER2_COMPB_vect) { // Обработчик по совпадению с регистром B
  if(!zFlag) return;
  PORTB &= ~_BV(PB0); // Формируем спад импульса на шине Zero
  zFlag = 0; // Запрещаем формирование "ложных" импульсов Zero
}
//------------------------------------------------------------------------------
int main(void) {
  DDRB = _BV(PB0); // Пин PB0 - на вывод (сигнал Zero)
  DDRB |= _BV(PB1); // Пин PB1 используется для диагностических целей
  // Инициализация TWI-модуля --------------------------------------------------
  TWAR = ADDR << 1;   // Адрес устройства - в старших 7 битах
  TWCR = _BV(TWIE);   // Разрешаем прерывания от TWI
  TWCR |= _BV(TWEA);  // Разрешаем подачу импульса ACK
  TWCR |= _BV(TWINT); // Очистим флаг TWINT (очищается при установке в 1)
  TWCR |= _BV(TWEN);  // Разрешаем работу TWI и активирует TWI-интерфейс
  // Инициализация таймера 0 для тактирования АЦП ------------------------------
  TCCR0A = _BV(WGM01); // Счетчик работает в режиме CTC (сброс по совпадению)
  TCCR0B = _BV(CS01) | _BV(CS00); // Предделитель на 64 (на счетчик - 250 кГц)
  OCR0A = T_ADC; // Определяет период запуска АЦП
  TIMSK0 |= _BV(OCIE0A); // Разрешаем прерывания по совпадению с OCR0A
  // Инициализация АЦП ---------------------------------------------------------
  ADMUX = 0; // Канал 0 (ADC0), MUX3...0 = 0
  ADMUX = _BV(REFS1) | _BV(REFS0); // Внутр. источник опорного напряжения 1.1V
  ADCSRA = _BV(ADEN); // АЦП включен
  ADCSRA |= _BV(ADIE); // По окончанию преобразования - прерывание
  ADCSRA |= _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0); // Предделитель на 128
  // Инициализация внешнего прерывания INT0
  EICRA = _BV(ISC01); // Прерывания по спаду уровня на пине PD2/INT0
  EIMSK = _BV(INT0); // Разрешаем внешние прерывания на пине PD2/INT0
  // Инициализация таймера 2 для формирования импульса нуля Zero
  // Счетчик работает в обычном режиме (Normal mode)
  TCCR2B |= _BV(CS22) | _BV(CS20); // Предделитель на 128 (сч. - 125 кГц)
  OCR2A = ZERO_START; // Фронт импульса Zero
  OCR2B = ZERO_START + ZERO_WIDTH; // Спад импульcа Zero
  TIMSK2 = _BV(OCIE2A) | _BV(OCIE2B); // Разрешаем прерывания по совпадению

  register uint32_t csum;

  sei(); // Разрешаем глобальные прерывания

  ptr = buf;
  for(int i = 0; i < BUF_SIZE; buf[i++] = 15);
  while(1) { // Бесконечный цикл
    if(prFlag) { // По шине i2c принят новый пакет - отрабатываем его
      cmd = buf[0]; // Код команды
      switch(cmd) { // Анализируем и выполняем команды
        case 0x33: // Чтение ROM - в первом байте - код семейства
          buf[0] = FAMILY_CODE;
          break;
        case 0x4e: // write scratch - записать данные в рабочую область
          memcpy(scratch, &buf[1], EE_SIZE);
          break;
        case 0xbe: // read scratch - считать данные из рабочей области
          memcpy(buf, scratch, EE_SIZE);
          break;
        case 0x48: // copy scratch - записать блок данных в EEPROM
          eeprom_update_block((const void *)scratch, (void *)0, EE_SIZE);
          break;
        case 0xb8: // recall EEPROM - считать блок данных в EEPROM
          eeprom_read_block((void *)scratch, (const void*)0, EE_SIZE);
          break;
        case 0x52: // Команда управления
          break;
        case 0x53: // Защелкнуть данные в буфер обмена (младший байт - первый)
          csum = crms;
          buf[0] = csum & 0xFF; csum >>= 8;
          buf[1] = csum & 0xFF; csum >>= 8;
          buf[2] = csum & 0xFF; csum >>= 8;
          buf[3] = csum & 0xFF;
          break;
      }
      prFlag = 0; // Сбросим флаг получения нового пакета
    }
  }
  return 0;
}
//------------------------------------------------------------------------------
// История и список изменений
// 1. 28.11.2018 - Отлажен весь функционал для этого типа устройств.
//------------------------------------------------------------------------------
