//------------------------------------------------------------------------------
// Прошивка для контроллера PDM
// Код семейства - 0xA3
// Адрес на шине i2c - 0x13
// OldBean, 28.11.2018
//------------------------------------------------------------------------------
#include <avr/io.h>
#include <util/twi.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <string.h>
//------------------------------------------------------------------------------
#define F_CPU 16000000UL // Тактовая частота МК
#include <util/delay.h>

#define ADDR 0x13        // Адрес устройства на шине I2C
#define FAMILY_CODE 0xA3 // Код семейства данного устройства
#define BUF_SIZE 32      // Размер буфера обмена данными
#define EE_SIZE 6        // Размер блока идентиф. и калибр. данных
//------------------------------------------------------------------------------
#define PULSE_WIDTH 78 // Длительность управляющего импульса - около 5 мс
static volatile uint8_t pdm = 0;     // Требуемый уровень PDM-модуляции в %
static volatile int16_t err = 0;     // Ошибка дискретизации
static volatile int8_t pp = 1;       // Относительный знак предстоящего
                                     // полупериода 1 - положительный, 0 - 
                                     // отрицательный
static volatile int16_t ps = 0;      // Отклонение постоянной составляющей от 0
static volatile uint8_t pFlag = 0;   // 1 - пропускаем полупериод к нагрузке,
                                     // т.е. подаем управляющий импульс на
                                     // моську, 0 - нет
static volatile int16_t lev;         // Текущий уровень (pdm + err)
//------------------------------------------------------------------------------
static uint8_t buf[BUF_SIZE];        // Буфер обмена
static uint8_t scratch[BUF_SIZE];    // Рабочая область RAM
static volatile uint8_t *ptr;        // Указатель на байт буфера обмена
static volatile uint8_t runFlag = 1; // Рабочий режим, 0 - режим настройки
static volatile uint8_t nb = 0;      // Счетчик байтов
static volatile uint8_t prFlag = 0;  // Принят новый пакет байтов
static volatile uint8_t cmd = 0x00;  // Код команды
//------------------------------------------------------------------------------
ISR(TWI_vect) { // Обработчик событий модуля TWI
  switch(TW_STATUS) {     // Статус - в 5 старших битах регистра TWSR
    // Принимаем байты
    case TW_SR_SLA_ACK:   // Адрес для приема принят, ACK отправлен - 0x80
      ptr = buf;          // Настраиваемся на начало буфера обмена
      nb = 0;             // Сброс счетчика байтов
      break;
    case TW_SR_DATA_ACK:  // Очередной байт принят, ACK отправлен - 0x80
      *ptr++ = TWDR;
      nb++;
      break;
    case TW_SR_STOP:      // Состояние "стоп" на шине - 0xA0
      prFlag = 1;         // Новый пакет принят
      break;
    // Отправляем байты
    case TW_ST_SLA_ACK:   // Адрес для отправки принят, ACK отправлен - 0xA8
      ptr = buf;          // Настраиваемся на начало буфера обмена
      nb = 0;             // Сброс счетчика байтов
    case TW_ST_DATA_ACK:  // Очередной байт отправлен, ACK принят - 0xB8
      TWDR = *ptr++;
      nb++;
      break;
    case TW_ST_DATA_NACK: // Байт отправлен и получен NACK - 0xC0
    case TW_ST_LAST_DATA: // Отправлен последний байт, получен ACK - 0xC8
      break;
    // Ошибки старта или стопа - сброс TWI
    case TW_BUS_ERROR:
      TWCR = _BV(TWIE) | _BV(TWINT) | _BV(TWEA) | _BV(TWEN);
      break;
  }
  TWCR |= _BV(TWINT); // Очистим флаг TWINT (очищается при установке в 1)
}
//------------------------------------------------------------------------------
ISR(INT0_vect) { // Внешнее прерывание INT0 (Zero от детектора нуля)
  lev = pdm + err; // Текущий уровень выходного сигнала (мощности)
  pFlag = (lev >= 50)? 1 : 0; // Предварительное решение
  if(ps*pFlag*pp > 0)         // Однако, если подадим импульс, то постоянная
    pFlag = 0;                // составляющая возрастет. Значит - не будем
                              // подавать - отложим да следующего полупериода
  if(pFlag) { // Положительное решение принято - дадим этот полупериод нагрузке
    PORTD |= _BV(PD0);        // Подаем управляющий импульс
    err = lev - 100;          // Оцениваем ошибку (сколько "переложили")
  } else err = lev;           // Если не дали - сколько "недоложили"
  GTCCR = _BV(PSRASY);        // Сброс предделителя таймера 2
  TCNT2 = 0;                  // Обнуляем счетчик таймера 2
}
//------------------------------------------------------------------------------
ISR(TIMER2_COMPA_vect) { // Обработчик по совпадению с регистром A
  PORTD &= ~_BV(PD0); // Завершаем управляющий импульс
}
//------------------------------------------------------------------------------
int main(void) {
  DDRD |= _BV(PD0);   // Пин 0 порта D - на вывод
  // Инициализация TWI-модуля --------------------------------------------------
  TWAR = ADDR << 1;   // Адрес устройства - в старших 7 битах
  TWCR = _BV(TWIE);   // Разрешаем прерывания от TWI
  TWCR |= _BV(TWEA);  // Разрешаем подачу импульса ACK
  TWCR |= _BV(TWINT); // Очистим флаг TWINT (очищается при установке в 1)
  TWCR |= _BV(TWEN);  // Разрешаем работу TWI и активирует TWI-интерфейс
  // Инициализация внешнего прерывания INT0
  EICRA = _BV(ISC01) | _BV(ISC00); // Прерывания по фронту на пине PD2/INT0
  EIMSK = _BV(INT0); // Разрешаем внешние прерывания на пине PD2/INT0
  // Инициализация таймера 2 для формирования управляющего импульса для
  // MOC3083 длительностью приблизительно 5 мс
  // Счетчик работает в обычном режиме (Normal mode)
  TCCR2B |= _BV(CS22) | _BV(CS21) | _BV(CS20); // Предделитель на 1024
  OCR2A = PULSE_WIDTH; // Длительность управляющего импульса
  TIMSK2 = _BV(OCIE2A); // Разрешаем прерывания по совпадению

  sei(); // Разрешаем глобальные прерывания

  ptr = buf;
  while(1) { // Бесконечный цикл
    if(prFlag) { // По шине i2c принят новый пакет - отрабатываем его
      cmd = buf[0]; // Код команды
      switch(cmd) { // Анализируем и выполняем команды
        case 0x33: // Чтение ROM - в первом байте - код семейства
          buf[0] = FAMILY_CODE;
          break;
        case 0x4e: // write scratch - записать данные в рабочую область
          memcpy(scratch, &buf[1], EE_SIZE);
          break;
        case 0xbe: // read scratch - считать данные из рабочей области
          memcpy(buf, scratch, EE_SIZE);
          break;
        case 0x48: // copy scratch - записать блок данных в EEPROM
          eeprom_update_block((const void *)scratch, (void *)0, EE_SIZE);
          break;
        case 0xb8: // recall EEPROM - считать блок данных в EEPROM
          eeprom_read_block((void *)scratch, (const void*)0, EE_SIZE);
          break;
        case 0x52: // Команда управления
          pdm = buf[1]; // Устанавливаем и корректируем уровень модуляции
          if(pdm > 100) pdm = 100;
          break;
      }
      prFlag = 0; // Сбросим флаг получения нового пакета
    }
  }
  return 0;
}
//------------------------------------------------------------------------------
// История и список изменений
// 1. 28.11.2018 - Отлажен весь функционал для этого типа устройств.
//------------------------------------------------------------------------------
